/*
 * This file derives from SFMT 1.3.3
 * (http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/SFMT/index.html), which was
 * released under the terms of the following license:
 *
 *   Copyright (c) 2006,2007 Mutsuo Saito, Makoto Matsumoto and Hiroshima
 *   University. All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are
 *   met:
 *
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above
 *         copyright notice, this list of conditions and the following
 *         disclaimer in the documentation and/or other materials provided
 *         with the distribution.
 *       * Neither the name of the Hiroshima University nor the names of
 *         its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written
 *         permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef SFMT_PARAMS2281_H
#define SFMT_PARAMS2281_H

#define POS1 12
#define SL1 19
#define SL2 1
#define SR1 5
#define SR2 1
#define MSK1 0xbff7ffbfU
#define MSK2 0xfdfffffeU
#define MSK3 0xf7ffef7fU
#define MSK4 0xf2f7cbbfU
#define PARITY1 0x00000001U
#define PARITY2 0x00000000U
#define PARITY3 0x00000000U
#define PARITY4 0x41dfa600U

/* PARAMETERS FOR ALTIVEC */
#if defined(__APPLE__) /* For OSX */
#	define ALTI_SL1 (vector unsigned int)(SL1, SL1, SL1, SL1)
#	define ALTI_SR1 (vector unsigned int)(SR1, SR1, SR1, SR1)
#	define ALTI_MSK (vector unsigned int)(MSK1, MSK2, MSK3, MSK4)
#	define ALTI_MSK64 (vector unsigned int)(MSK2, MSK1, MSK4, MSK3)
#	define ALTI_SL2_PERM                                                  \
		(vector unsigned char)(1, 2, 3, 23, 5, 6, 7, 0, 9, 10, 11, 4,  \
		    13, 14, 15, 8)
#	define ALTI_SL2_PERM64                                                \
		(vector unsigned char)(1, 2, 3, 4, 5, 6, 7, 31, 9, 10, 11, 12, \
		    13, 14, 15, 0)
#	define ALTI_SR2_PERM                                                  \
		(vector unsigned char)(7, 0, 1, 2, 11, 4, 5, 6, 15, 8, 9, 10,  \
		    17, 12, 13, 14)
#	define ALTI_SR2_PERM64                                                \
		(vector unsigned char)(15, 0, 1, 2, 3, 4, 5, 6, 17, 8, 9, 10,  \
		    11, 12, 13, 14)
#else /* For OTHER OSs(Linux?) */
#	define ALTI_SL1                                                       \
		{ SL1, SL1, SL1, SL1 }
#	define ALTI_SR1                                                       \
		{ SR1, SR1, SR1, SR1 }
#	define ALTI_MSK                                                       \
		{ MSK1, MSK2, MSK3, MSK4 }
#	define ALTI_MSK64                                                     \
		{ MSK2, MSK1, MSK4, MSK3 }
#	define ALTI_SL2_PERM                                                  \
		{ 1, 2, 3, 23, 5, 6, 7, 0, 9, 10, 11, 4, 13, 14, 15, 8 }
#	define ALTI_SL2_PERM64                                                \
		{ 1, 2, 3, 4, 5, 6, 7, 31, 9, 10, 11, 12, 13, 14, 15, 0 }
#	define ALTI_SR2_PERM                                                  \
		{ 7, 0, 1, 2, 11, 4, 5, 6, 15, 8, 9, 10, 17, 12, 13, 14 }
#	define ALTI_SR2_PERM64                                                \
		{ 15, 0, 1, 2, 3, 4, 5, 6, 17, 8, 9, 10, 11, 12, 13, 14 }
#endif /* For OSX */
#define IDSTR "SFMT-2281:12-19-1-5-1:bff7ffbf-fdfffffe-f7ffef7f-f2f7cbbf"

#endif /* SFMT_PARAMS2281_H */
