#ifndef JEMALLOC_INTERNAL_OVERRIDES_H
#define JEMALLOC_INTERNAL_OVERRIDES_H

/*
 * Under normal circumstances this header serves no purpose, as these settings
 * can be customized via the corresponding autoconf options at configure-time.
 * Overriding in this fashion is useful when the header files generated by
 * autoconf are used as input for another build system.
 */

#ifdef JEMALLOC_OVERRIDE_LG_PAGE
#	undef LG_PAGE
#	define LG_PAGE JEMALLOC_OVERRIDE_LG_PAGE
#endif

#ifdef JEMALLOC_OVERRIDE_JEMALLOC_CONFIG_MALLOC_CONF
#	undef JEMALLOC_CONFIG_MALLOC_CONF
#	define JEMALLOC_CONFIG_MALLOC_CONF                                    \
		JEMALLOC_OVERRIDE_JEMALLOC_CONFIG_MALLOC_CONF
#endif

#endif /* JEMALLOC_INTERNAL_OVERRIDES_H */
